#include "GLCam.h"

#include <SDL/SDL_opengl.h>
#include <cmath>
#include <iostream>

#ifndef M_PI
#define M_PI (REAL_PI)
#endif

using namespace ambenv;

GLCam::GLCam( const Vec3 & eye, const Vec3 & lookAt, const Vec3 & up )
  : m_eye( eye ), m_lookAt( lookAt ), m_up( up )
{
  m_lookAtHistory.resize( SMOOTHING_FRAMES );
}

void GLCam::load() const {

  gluLookAt( m_eye.x(), m_eye.y(), m_eye.z(),
     m_lookAt.x(), m_lookAt.y(), m_lookAt.z(),
     m_up.x(), m_up.y(), m_up.z() );
}

void GLCam::setEye( const Vec3 & eye ) 
{
  m_eye = eye;
}

void GLCam::setLookAt( const Vec3 & lookAt, bool forceNow ) 
{
  Vec3 newLookAt;

  if( forceNow ) {
    newLookAt = lookAt;
  } else {
    newLookAt = smoothCamera( lookAt );
  }
  
  m_lookAt = newLookAt;
}

void GLCam::orbitXZ( Real angle )
{
  Vec3 D = m_eye - m_lookAt;

  double ang = (M_PI/180.0) * angle;
  double newx = m_lookAt.x() + (D.x() * cos(ang) - D.z() * sin(ang));
  double newz = m_lookAt.z() + (D.x() * sin(ang) + D.z() * cos(ang));

  m_eye.x( (Real)newx );
  m_eye.z( (Real)newz );
}

void GLCam::orbitXY( Real angle ) 
{
  Vec3 D = m_eye - m_lookAt;

  Real norm = (Real)sqrt( D.x() * D.x() + D.z() * D.z() );
  Real ang = (M_PI/180.0) * angle;
  Real newnorm = norm * cos(ang) - D.y() * sin( ang );
  Real newy = m_lookAt.y() + (norm * (Real)sin(ang) + D.y() * (Real)cos(ang));
  Real newx = m_lookAt.x() + D.x() * newnorm / norm;
  Real newz = m_lookAt.z() + D.z() * newnorm / norm;
  
  if( newnorm > 0 ) {
    m_eye = Vec3( newx, newy, newz );
  }
}

void GLCam::rotateXZ( Real angle ) 
{
  double xD = m_lookAt.x() - m_eye.x();
  double zD = m_lookAt.z() - m_eye.z();

  double currentAngle = atan2( zD, xD );
  double newAngle = currentAngle + (angle * M_PI / 180.0);

  double newxD = xD * cos( newAngle ) / cos( currentAngle );
  double newzD = zD * sin( newAngle ) / sin( currentAngle );

  newxD += m_eye.x();
  newzD += m_eye.z();

  m_lookAt.x( (Real)newxD );
  m_lookAt.z( (Real)newzD );
}

void GLCam::rotateXY(float angle) {
  /*
double xD = m_aLook[0] - m_aEye[0];
double yD = m_aLook[1] - m_aEye[1];
double zD = m_aLook[2] - m_aEye[2];

double radius = sqrt(xD * xD + yD * yD + zD * zD);
double xzRadius = sqrt(xD * xD + zD * zD);

double currentAngle = atan2(yD, xzRadius);

double newAngle = currentAngle + (angle * PI / 180.0);
if(newAngle >= (89.0 * PI / 180.0)) {
	newAngle = 89.0 * PI / 180.0;
}
if(newAngle <= (-89.0 * PI / 180.0)) {
	newAngle = -89.0 * PI / 180.0;
}

double newyD = radius * sin(newAngle);
double newXZRadius = xzRadius * cos(newAngle) / cos(currentAngle);
double newxD = xD * newXZRadius / xzRadius;
double newzD = zD * newXZRadius / xzRadius;

newxD += m_aEye[0];
newyD += m_aEye[1];
newzD += m_aEye[2];

m_aLook[0] = (float)newxD;
m_aLook[1] = (float)newyD;
m_aLook[2] = (float)newzD;
  */
}

void GLCam::dolly( Real amount ) 
{

  Vec3 D = m_eye - m_lookAt;

  double radius = D.length();

  double newAmount = amount * radius / 100.0f;

  if( fabs( newAmount ) < fabs( amount / 4.0 ) ) {
    newAmount = amount / 4.0;
  }
  
  double newnorm = radius + newAmount;

  if( newnorm < 0.02f ) {
    newnorm = 0.02f;
  }

  m_eye = D * (newnorm / radius) + m_lookAt;

}

void GLCam::translate( Real mx, Real my ) 
{
  Vec3 v = m_eye - m_lookAt;
  const static Vec3 psuedoUp( 0.0, 1.0, 0.0 );

  Vec3 tu = v.cross( psuedoUp );
  tu.normalize();

  Vec3 tv = tu.cross( v );
  tv.normalize();


  // -- scale tu, tv by translate amounts
  tu *= mx;
  tv *= my;


  m_eye += (tu + tv);
  m_lookAt += (tu + tv);
}

void GLCam::updateInverse() {

//double mat[16];
//glGetDoublev(GL_MODELVIEW_MATRIX, mat);
//invertMatrix(m_aCamInv, mat);
}

double* GLCam::getInverse() {

  return m_aCamInv;
}



Vec3 GLCam::smoothCamera( const Vec3 & lookAt ) 
{
  Real weight = SMOOTHING_WEIGHT;

  Vec3 newLookAt;
  {
  // -- move lookAt down
  for( int i=0; i<((int)m_lookAtHistory.size() - 1); i++ ) {
    m_lookAtHistory[i] = m_lookAtHistory[i+1];
  }
  }

  // -- add current lookAt to tail of array
  *(m_lookAtHistory.end()) = lookAt;
  {
  // -- starting at tail, add lookAt, weighting and adjusting appropriately
  for( int i=((int)m_lookAtHistory.size() - 1); i>=0; i-- ) {
    newLookAt += m_lookAtHistory[i] * weight;
    weight *= SMOOTHING_FACTOR;
  }
  }
  
  return( newLookAt );
}

void GLCam::drawShadowBox() 
{
  /*
  glEnable( GL_BLEND );
  glDisable( GL_LIGHTING );
  glDisable( GL_CULL_FACE );

glPushMatrix();

glTranslatef(m_aEye[0], m_aEye[1], m_aEye[2]);

const float BOX_SIZE = 10.0f;

glBegin(GL_QUADS);

glColor4f(0.0, 0.0, 0.0, 0.2f);
glVertex3f(-BOX_SIZE, BOX_SIZE, -BOX_SIZE);
glVertex3f(-BOX_SIZE, BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE, BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE, BOX_SIZE, -BOX_SIZE);

glVertex3f( BOX_SIZE, -BOX_SIZE, -BOX_SIZE);
glVertex3f( BOX_SIZE, -BOX_SIZE,  BOX_SIZE);
glVertex3f(-BOX_SIZE, -BOX_SIZE,  BOX_SIZE);
glVertex3f(-BOX_SIZE, -BOX_SIZE, -BOX_SIZE);

glVertex3f(-BOX_SIZE, BOX_SIZE, -BOX_SIZE);
glVertex3f(-BOX_SIZE,-BOX_SIZE, -BOX_SIZE);
glVertex3f(-BOX_SIZE,-BOX_SIZE,  BOX_SIZE);
glVertex3f(-BOX_SIZE, BOX_SIZE,  BOX_SIZE);

glVertex3f( BOX_SIZE, BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE,-BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE,-BOX_SIZE, -BOX_SIZE);
glVertex3f( BOX_SIZE, BOX_SIZE, -BOX_SIZE);

glVertex3f(-BOX_SIZE, BOX_SIZE, -BOX_SIZE);
glVertex3f(-BOX_SIZE,-BOX_SIZE, -BOX_SIZE);
glVertex3f( BOX_SIZE,-BOX_SIZE, -BOX_SIZE);
glVertex3f( BOX_SIZE, BOX_SIZE, -BOX_SIZE);

glVertex3f(-BOX_SIZE,-BOX_SIZE,  BOX_SIZE);
glVertex3f(-BOX_SIZE, BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE, BOX_SIZE,  BOX_SIZE);
glVertex3f( BOX_SIZE,-BOX_SIZE,  BOX_SIZE);

glEnd();

glPopMatrix();

glEnable(GL_CULL_FACE);
glEnable(GL_LIGHTING);
glDisable(GL_BLEND);
  */
}
