#include "Hemicube.h"

using namespace ambenv;

#ifndef M_PI
#define M_PI (REAL_PI)
#endif

const Real nearClip = 1.0;
const Real farClip = 1000.0;

Hemicube::Hemicube( unsigned size ) : m_size( size )
{
}

void Hemicube::renderScene( const MeshVertex & vertex, const Mesh * mesh,
			    render_scene_callback callback )
{
  
  const int incr1 = m_size / 4;
  const int incr2 = 2*incr1;
  const int incr3 = 3*incr1;

  Vec3 lookAt;
  Vec3 n = vertex.n();
  Vec3 up( 0, 1, 0 );
  if( n.approximates( up ) || n.approximates( up * -1 ) ) {
    up = Vec3( 0, 0, 1 );
  }
  
  glClear( GL_COLOR_BUFFER_BIT );
  
  glColor3f( 1.0, 1.0, 1.0 );
  // -- Center
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( -1, 1, -1, 1, nearClip, farClip );
  glViewport( incr1, incr1, incr2, incr2 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  
  lookAt = vertex.v() + n;
  gluLookAt( vertex.v().x(), vertex.v().y(), vertex.v().z(),
	     lookAt.x(), lookAt.y(), lookAt.z(),
	     up.x(), up.y(), up.z() );
  callback( mesh );
  
  
  glColor3f( 1.0, 0.0, 0.0 );
  // -- Right
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( 0, 1, -1, 1, nearClip, farClip );
  glViewport( 0, incr1, incr1, incr2 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();      
  
  Vec3 right = up.cross( n );
  
  lookAt = vertex.v() + right;
  gluLookAt( vertex.v().x(), vertex.v().y(), vertex.v().z(),
	     lookAt.x(), lookAt.y(), lookAt.z(),
	     up.x(), up.y(), up.z() );
  
  callback( mesh );
  
  
  glColor3f( 0.0, 1.0, 0.0 );
  // -- Left
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( -1, 0, -1, 1, nearClip, farClip );
  glViewport( incr3, incr1, incr1, incr2 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  
  lookAt = vertex.v() - right;
  gluLookAt( vertex.v().x(), vertex.v().y(), vertex.v().z(),
	     lookAt.x(), lookAt.y(), lookAt.z(),
	     up.x(), up.y(), up.z() );
  
  callback( mesh );
  
  
  glColor3f( 0.0, 0.0, 1.0 );
  // -- Back
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( -1, 1, 0, 1, nearClip, farClip );
  glViewport( incr1, 0, incr2, incr1 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  
  lookAt = vertex.v() + up;
  gluLookAt( vertex.v().x(), vertex.v().y(), vertex.v().z(),
	     lookAt.x(), lookAt.y(), lookAt.z(),
	     -n.x(), -n.y(), -n.z() );  
  
  callback( mesh );
  
  
  glColor3f( 1.0, 1.0, 0.0 );
  // -- Forward
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( -1, 1, -1, 0, nearClip, farClip );
  glViewport( incr1, incr3, incr2, incr1 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  
  lookAt = vertex.v() - up;
  gluLookAt( vertex.v().x(), vertex.v().y(), vertex.v().z(),
	     lookAt.x(), lookAt.y(), lookAt.z(),
	     n.x(), n.y(), n.z() );  
  
  callback( mesh );
}

//! See "http://www.fsz.bme.hu/~szirmay/radiosit/node7.html" for reference.
/*!
* -- w(x,y,z) := y / PI * (x^2 + y^2 + z^2)^2
*/  
void Hemicube::computeWeightsAndNormals( std::vector<Real> & weights, 
                                         std::vector<Vec3> & normals )
{

  const int quarter1 = m_size / 4;
  const int quarter3 = quarter1 * 3;

  weights.resize( m_size * m_size );
  normals.resize( m_size * m_size ); 

  Real weight;
  Vec3 pixel;

  Real incr = 1.0 / (Real)quarter1;
  
  // -- Format the weights like the viewports in the framebuffer.
  
  // -- Front: (i,j) pixel corresponds to x=[-1,1], y=[0,1], z= 1 in world space.
  // -- Back:  (i,j) pixel corresponds to x=[-1,1], y=[0,1], z=-1 in world space.
  pixel = Vec3( -1, 0, 1 );
  {
    for( int i=0; i<quarter1; i++ ) {

      pixel.y( pixel.y() + incr );
      pixel.x( -1 );
      
      for( int j=quarter1; j<quarter3; j++ ) {
        
	pixel.x( pixel.x() + incr );

        weight = pixel.x() * pixel.x() + pixel.y() * pixel.y() + 1;
        weight = M_PI * weight * weight;
        weight = pixel.y() / weight;


        weights[m_size*i + j] = weight;
        normals[m_size*i + j] = pixel;
	normals[m_size*i + j].normalize();


	Vec3 opp_pixel = pixel;
	opp_pixel.z( -1 );
	opp_pixel.normalize();
        
        weights[m_size*(m_size - 1 - i) + j] = weight;
        normals[m_size*(m_size - 1 - i) + j] = opp_pixel;
      }
    }
  }
  

  // -- Right: (i,j) pixel corresponds to y=[0,1], z=[-1, 1], x=-1 in world space.
  // -- Left:  (i,j) pixel corresponds to y=[0,1], z=[-1, 1], x= 1 in world space.
  pixel = Vec3( -1, 1, -1 );
  {
    for( int i=quarter1; i<quarter3; i++ ) {
      
      pixel.y( 0 );
      pixel.z( pixel.z() + incr );

      for( int j=0; j<quarter1; j++ ) {
        
	pixel.y( pixel.y() + incr );

        weight = 1 + pixel.y() * pixel.y() + pixel.z() * pixel.z();
        weight = M_PI * weight * weight;
        weight = pixel.y() / weight;

        weights[m_size*i + j] = weight; 
        normals[m_size*i + j] = pixel;
	normals[m_size*i + j].normalize();
        
        Vec3 opp_pixel = pixel;
        opp_pixel.x( 1 );
	opp_pixel.normalize();

        weights[m_size*i + (m_size - 1 - j)] = weight; 
        normals[m_size*i + (m_size - 1 - j)] = opp_pixel;
      } 
    }
  }
  

  // -- Center: (i,j) pixels cooresponds to z=[1,-1], x=[-1,1], y=1 in world space.
  pixel = Vec3( -1, 1, 1 );
  {
    for( int i=quarter1; i<quarter3; i++ ) {
      
      pixel.z( 1 );
      pixel.x( pixel.x() + incr );

      for( int j=quarter1; j<quarter3; j++ ) {  
        
	pixel.z( pixel.z() - incr );

        weight = pixel.x() * pixel.x() + pixel.z() * pixel.z() + 1;
        weight = M_PI * weight * weight;
        weight = 1.0 / weight;

        weights[m_size*i + j] = weight;
        normals[m_size*i + j] = pixel;
	normals[m_size*i + j].normalize();
      }
    }
  }
  
}
