#include "Vec3.h"

#include <cassert>
#include <math.h>

using namespace ambenv;

Vec3::Vec3() : m_x( 0.0f ), m_y( 0.0f ), m_z( 0.0f )
{
}

Vec3::Vec3( Real x, Real y, Real z ) : m_x( x ), m_y( y ), m_z( z )
{
}

Vec3::Vec3( const float * xyz ) : m_x( xyz[0] ), m_y( xyz[1] ), m_z( xyz[2] )
{
}

Vec3::Vec3(const double* xyz) : m_x( xyz[0] ), m_y( xyz[1] ), m_z( xyz[2] )
{
}

// -- accessors
Real Vec3::x() const
{
  return( m_x );
}
Real Vec3::y() const
{
  return( m_y );
}
Real Vec3::z() const
{
  return( m_z );
}
void Vec3::x( Real x )
{
  m_x = x;
}
void Vec3::y( Real y )
{
  m_y = y;
}
void Vec3::z( Real z )
{
  m_z = z;
}


// -- vector operations
Vec3 Vec3::cross( const Vec3 & rhs ) const
{
  return( Vec3( m_y * rhs.z() - m_z * rhs.y(),
		m_z * rhs.x() - m_x * rhs.z(),
		m_x * rhs.y() - m_y * rhs.x() ) );
}

Real Vec3::dot( const Vec3 & rhs ) const
{
  return( m_x * rhs.x() + m_y * rhs.y() + m_z * rhs.z() );
}

void Vec3::normalize()
{
  Real mag = length();
  (*this) /= mag;
}

Real Vec3::length() const
{
  return( (Real)sqrt( lengthSq() ) );
}

Real Vec3::lengthSq() const
{
  return( m_x*m_x + m_y*m_y + m_z*m_z );
}

Real Vec3::sum() const
{
  return( m_x + m_y + m_z );
}

Real Vec3::average( const Vec3 & weights ) const
{
  return( (m_x*weights.x() + m_y*weights.y() + m_z*weights.z()) / weights.sum() );
}

bool Vec3::approximates( const Vec3 & rhs ) const
{
  if( (fabs( m_x - rhs.x() ) > EPSILON) ||
      (fabs( m_y - rhs.y() ) > EPSILON) ||
      (fabs( m_z - rhs.z() ) > EPSILON) ) {
    return( false );
  } else {
    return( true );
  }
}


bool Vec3::operator==( const Vec3 & rhs ) const
{
  if( (m_x != rhs.x()) || (m_y != rhs.y()) || (m_z != rhs.z()) ) {
    return( false );
  } else {
    return( true );
  }
}

bool Vec3::operator!=( const Vec3 & rhs ) const
{
  return( !(this->operator==( rhs ) ) );
}


// -- point-wise operators
Vec3 Vec3::power( Real exponent ) const
{
  Vec3 result( (Real)pow( m_x, exponent ),
	       (Real)pow( m_y, exponent ),
	       (Real)pow( m_z, exponent ) );
  return( result );
}


Vec3 Vec3::operator+( const Vec3 & rhs ) const
{
  return( Vec3( m_x + rhs.x(),
		m_y + rhs.y(),
		m_z + rhs.z() ) );
}
Vec3 Vec3::operator+( Real rhs ) const
{
  return( Vec3( m_x + rhs,
		m_y + rhs,
		m_z + rhs ) );
}
void Vec3::operator+=( const Vec3 & rhs )
{
  m_x += rhs.x();
  m_y += rhs.y();
  m_z += rhs.z();
}
void Vec3::operator+=( Real rhs )
{
  m_x += rhs;
  m_y += rhs;
  m_z += rhs;
}

Vec3 Vec3::operator-( const Vec3 & rhs ) const
{
  return( (*this) + ( rhs * -1.0 ) );
}
Vec3 Vec3::operator-( Real rhs ) const
{
  return( (*this) + ( rhs * -1.0 ) );
}
void Vec3::operator-=( const Vec3 & rhs )
{
  (*this) += ( rhs * -1.0 );
}
void Vec3::operator-=( Real rhs )
{
  (*this) += ( rhs * -1.0 );
}

Vec3 Vec3::operator*( const Vec3 & rhs ) const
{
  return( Vec3( m_x * rhs.x(),
                m_y * rhs.y(),
                m_z * rhs.z() ) );
}
Vec3 Vec3::operator*( Real rhs ) const
{
  return( Vec3( m_x * rhs,
                m_y * rhs,
                m_z * rhs ) );
}
void Vec3::operator*=( const Vec3 & rhs )
{
  m_x *= rhs.x();
  m_y *= rhs.y();
  m_z *= rhs.z();
}
void Vec3::operator*=( Real rhs )
{
  m_x *= rhs;
  m_y *= rhs;
  m_z *= rhs;
}

Vec3 Vec3::operator/( const Vec3 & rhs ) const
{
  assert( rhs.x() != 0.0 );
  assert( rhs.y() != 0.0 );
  assert( rhs.z() != 0.0 );

  return( Vec3( m_x / rhs.x(),
                m_y / rhs.y(),
                m_z / rhs.z() ) );
}
Vec3 Vec3::operator/( Real rhs ) const
{
  assert( rhs != 0.0 );

  return( (*this) * ( 1.0 / rhs ) );
}
void Vec3::operator/=( const Vec3 & rhs )
{
  assert( rhs.x() != 0.0 );
  assert( rhs.y() != 0.0 );
  assert( rhs.z() != 0.0 );

  m_x /= rhs.x();
  m_y /= rhs.y();
  m_z /= rhs.z();
}
void Vec3::operator/=( Real rhs )
{
  assert( rhs != 0.0 );

  (*this) *= ( 1.0 / rhs );
}
